from threading import Thread
import queue
from . import packet_util
from .message_consumer import MessageConsumer
import logging


class MessageWorker(Thread):
    def __init__(self, socket, room_id):
        Thread.__init__(self)
        self.need_stop = False
        self.socket = socket
        self.room_id = room_id
        self.msg_queue = queue.Queue()
        self.message_consumer = MessageConsumer(self.msg_queue)

    def add_handler(self, msg_type, handler):
        self.message_consumer.add_handler(msg_type, handler)

    def set_stop(self, need_stop=True):
        self.need_stop = need_stop
        self.message_consumer.set_stop(need_stop)

    def prepare(self):
        self.message_consumer.start()
        self.enter_room()

    def enter_room(self):
        ori_str = packet_util.assemble_login_str(self.room_id)
        data = packet_util.assemble_transfer_data(ori_str)
        self.socket.send(data)
        ori_str = packet_util.assemble_join_group_str(self.room_id)
        data = packet_util.assemble_transfer_data(ori_str)
        self.socket.send(data)

    def run(self):
        self.prepare()
        while not self.need_stop:
            packet_size = self.socket.receive(4)
            if packet_size is None:
                logging.warning("Socket closed")
                self.socket.connect()
                self.enter_room()
                continue
            packet_size = int.from_bytes(packet_size, byteorder='little')
            data = self.socket.receive(packet_size)
            if data is None:
                logging.warning("Socket closed")
                self.socket.connect()
                self.enter_room()
                continue
            self.msg_queue.put(data)



from threading import Thread
import time
from . import packet_util


class HeartbeatWorker(Thread):
    def __init__(self, socket, heartbeat_interval=45):
        Thread.__init__(self)
        self.need_stop = False
        self.socket = socket
        self.heartbeat_interval = heartbeat_interval

    def set_stop(self, need_stop=True):
        self.need_stop = need_stop

    def run(self):
        while not self.need_stop:
            ori_str = packet_util.assemble_heartbeat_str()
            data = packet_util.assemble_transfer_data(ori_str)
            self.socket.send(data)
            time.sleep(self.heartbeat_interval)



import time

client_msg_type = 689
reserved_data_field = 0


def assemble_login_str(room_id):
    res = "type@=loginreq/roomid@=" + str(room_id) + "/"
    return res


def assemble_join_group_str(room_id):
    res = "type@=joingroup/rid@=" + str(room_id) + "/gid@=-9999/";
    return res


def assemble_heartbeat_str():
    res = "type@=keeplive/tick@=%s/" % int(time.time()) + "/"
    return res


def assemble_transfer_data(ori_str):
    data_size = len(ori_str)
    packet_size = 4 * 2 + data_size + 1;
    data = packet_size.to_bytes(4, byteorder='little')
    data += packet_size.to_bytes(4, byteorder='little')
    data += client_msg_type.to_bytes(2, byteorder='little')
    data += reserved_data_field.to_bytes(2, byteorder='little')
    data += ori_str.encode()
    data += b'\0'
    return data


def extract_str_from_data(data):
    packet_size = int.from_bytes(data[0:4], byteorder='little')
    if packet_size != len(data):
        return ""
    return data[8:].decode("utf8","ignore")


def parse_str_to_dict(ori_str):
    res = {}
    ori_strs = ori_str.split("/");
    for ori_str in ori_strs:
        kv = ori_str.split("@=")
        if len(kv) == 2:
            res[kv[0]] = kv[1]
    return res






from threading import Thread
from . import packet_util
import logging


class MessageConsumer(Thread):
    def __init__(self, msg_queue):
        Thread.__init__(self)
        self.need_stop = False
        self.msg_queue = msg_queue
        self.handlers = {}

    def add_handler(self, msg_type, handler):
        if msg_type not in self.handlers:
            self.handlers[msg_type] = []
        self.handlers[msg_type].append(handler)

    def set_stop(self, need_stop=True):
        self.need_stop = need_stop

    def run(self):
        while not self.need_stop:
            data = self.msg_queue.get()
            ori_str = packet_util.extract_str_from_data(data)
            msg = packet_util.parse_str_to_dict(ori_str)
            try:
                msg_type = msg['type']
                if msg_type in self.handlers:
                    for handler in self.handlers[msg_type]:
                        handler(msg)
            except Exception as e:
                logging.warning("Invalid msg received. Exception: %s"
                                % e)
            self.msg_queue.task_done()




import socket
import logging
import time


class TCPSocket(object):
    def __init__(self, host, port):
        self.host = host
        self.port = port
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.closed = True

    def send(self, data):
        if self.closed:
            return
        return self.socket.sendall(data)

    def close(self):
        if not self.closed:
            self.socket.close()
            self.closed = True

    def connect(self):
        if self.closed:
            while True:
                try:
                    self.socket.connect((self.host, self.port))
                except Exception as e:
                    logging.warning("Socket connect failed with %s:%d. Exception: %s"
                          % (self.host, self.port, e))
                    logging.warning("Try reconnect in 5 seconds")
                    time.sleep(5)
                    continue
                break
            self.closed = False

    def receive(self, target_size):
        data = b''
        while target_size:
            tmp = self.socket.recv(target_size)
            if not tmp:
                self.closed = True
                return None
            target_size -= len(tmp)
            data += tmp
        return data



from .message_worker import MessageWorker
from .heartbeat_worker import HeartbeatWorker
from .tcp_socket import TCPSocket


class Client(object):
    def __init__(self, room_id=562590, heartbeat_interval=45,
                 barrage_host="openbarrage.douyutv.com",
                 barrage_port=8601):
        self.room_id = room_id
        self.heartbeat_interval = heartbeat_interval
        self.barrage_host = barrage_host
        self.barrage_port = barrage_port
        self.tcp_socket = TCPSocket(self.barrage_host, self.barrage_port)
        self.message_worker = MessageWorker(self.tcp_socket, self.room_id)
        self.heartbeat_worker = HeartbeatWorker(self.tcp_socket, self.heartbeat_interval)

    def add_handler(self, msg_type, handler):
        self.message_worker.add_handler(msg_type, handler)

    def set_heartbeat_interval(self, heartbeat_interval):
        self.heartbeat_interval = heartbeat_interval

    def refresh_object(self):
        self.tcp_socket = TCPSocket(self.barrage_host, self.barrage_port)
        self.message_worker = MessageWorker(self.tcp_socket, self.room_id)
        self.heartbeat_worker = HeartbeatWorker(self.tcp_socket, self.heartbeat_interval)

    def set_room_id(self, room_id):
        self.room_id = room_id

    def start(self):
        self.tcp_socket.connect()
        self.message_worker.start()
        self.heartbeat_worker.start()

    def stop(self):
        self.message_worker.set_stop()
        self.heartbeat_worker.set_stop()
        self.tcp_socket.close()
        self.refresh_object()

